import numpy as np
import spacy
from spacy.tokens import Doc, Span, Token

from .base_parser import BaseParser, PTB_TOKEN_ESCAPE

__all__ = ['BeneparComponent', 'NonConstituentException']

# None is not allowed as a default extension value!
NOT_PARSED_SENTINEL = object()
Doc.set_extension('_constituent_data', default=NOT_PARSED_SENTINEL)

class NonConstituentException(Exception):
    pass

#%%
class ConstituentData():
    def __init__(self, starts, ends, labels, loc_to_constituent, label_vocab):
        self.starts = starts
        self.ends = ends
        self.labels = labels
        self.loc_to_constituent = loc_to_constituent
        self.label_vocab = label_vocab

class PartialConstituentData():
    def __init__(self):
        self.starts = [np.array([], dtype=int)]
        self.ends = [np.array([], dtype=int)]
        self.labels = [np.array([], dtype=int)]

    def finalize(self, doc, label_vocab):
        self.starts = np.hstack(self.starts)
        self.ends = np.hstack(self.ends)
        self.labels = np.hstack(self.labels)

        # TODO(nikita): Python for loops aren't very fast
        loc_to_constituent = np.full(len(doc), -1, dtype=int)
        prev = None
        for position in range(self.starts.shape[0]):
            if self.starts[position] != prev:
                prev = self.starts[position]
                loc_to_constituent[self.starts[position]] = position

        return ConstituentData(self.starts, self.ends, self.labels, loc_to_constituent, label_vocab)

#%%
class BeneparComponent(BaseParser):
    """
    Berkeley Neural Parser (benepar) component for spaCy.

    Sample usage:
    >>> nlp = spacy.load('en')
    >>> nlp.add_pipe(BeneparComponent("benepar_en"))
    >>> doc = nlp(u"The quick brown fox jumps over the lazy dog.")
    >>> sent = list(doc.sents)[0]
    >>> print(sent._.parse_string)

    This component is only responsible for constituency parsing. It should be
    preceded in the pipeline by other components that can perform sentence
    segmentation, tokenization, and part-of-speech tagging.
    """

    name = 'benepar'

    def __init__(self, filename, batch_size=64, disable_tagger=False):
        """
        Load a parsing model given a short model name (e.g. "benepar_en") or a
        filename on disk.

        name (str): Model name, or path to uncompressed TensorFlow model graph
        batch_size (int): Maximum number of sentences to process per batch
        disable_tagger (bool, default False): Unless disabled, the parser will
            set predicted part-of-speech tags for the document, overwriting any
            existing tags provided by spaCy models or previous pipeline steps.
            This option has no effect for older parser models that do not have
            a tagger built in.
        """
        super(BeneparComponent, self).__init__(filename, batch_size)
        self._do_tagging = (self._provides_tags and not disable_tagger)

    def __call__(self, doc):
        constituent_data = PartialConstituentData()
        for parse_raw, tags_raw, sent in self._batched_parsed_raw(self._process_doc(doc)):
            # The optimized cython decoder implementation doesn't actually
            # generate trees, only scores and span indices. Indices follow a
            # preorder traversal, which is also the order the ConstituentData
            # class requires.
            score, p_i, p_j, p_label = parse_raw

            # To remove null-labelled constituents that are used for
            # binarization, but not null-labeled terminal nodes
            valid = (p_label != 0) | (p_i + 1 == p_j)

            constituent_data.starts.append(p_i[valid] + sent.start)
            constituent_data.ends.append(p_j[valid] + sent.start)
            constituent_data.labels.append(p_label[valid])

            if self._do_tagging:
                for i, tag_idx in enumerate(tags_raw):
                    sent[i].tag_ = self._tag_vocab[tag_idx]

        doc._._constituent_data = constituent_data.finalize(doc, self._label_vocab)
        return doc

    def _process_doc(self, doc):
        for sent in doc.sents:
            yield [token.text for token in sent], sent

#%%

def get_constituent(span):
    constituent_data = span.doc._._constituent_data
    if constituent_data is NOT_PARSED_SENTINEL:
        raise Exception("No constituency parse is available for this document. Consider adding a BeneparComponent to the pipeline.")

    search_start = constituent_data.loc_to_constituent[span.start]
    if span.start + 1 < len(constituent_data.loc_to_constituent):
        search_end = constituent_data.loc_to_constituent[span.start + 1]
    else:
        search_end = len(constituent_data.ends)
    found_position = None
    for position in range(search_start, search_end):
        if constituent_data.ends[position] <= span.end:
            if constituent_data.ends[position] == span.end:
                found_position = position
            break

    if found_position is None:
        raise NonConstituentException("Span is not a constituent: {}".format(span))
    return constituent_data, found_position

def get_labels(span):
    constituent_data, position = get_constituent(span)
    label_num = constituent_data.labels[position]
    return constituent_data.label_vocab[label_num]

def parse_string(span):
    constituent_data, position = get_constituent(span)
    label_vocab = constituent_data.label_vocab
    doc = span.doc

    # Python 2 doesn't support "nonlocal", so wrap idx in a list
    idx_cell = [position - 1]
    def make_str():
        idx_cell[0] += 1
        i, j, label_idx = constituent_data.starts[idx_cell[0]], constituent_data.ends[idx_cell[0]], constituent_data.labels[idx_cell[0]]
        label = label_vocab[label_idx]
        if (i + 1) >= j:
            token = doc[i]
            s = u"({} {})".format(token.tag_, PTB_TOKEN_ESCAPE.get(token.text, token.text))
        else:
            children = []
            while ((idx_cell[0] + 1) < len(constituent_data.starts)
                and i <= constituent_data.starts[idx_cell[0] + 1]
                and constituent_data.ends[idx_cell[0] + 1] <= j):
                children.append(make_str())

            s = u" ".join(children)

        for sublabel in reversed(label):
            s = u"({} {})".format(sublabel, s)
        return s

    return make_str()

def get_subconstituents(span):
    constituent_data, position = get_constituent(span)
    label_vocab = constituent_data.label_vocab
    doc = span.doc

    while position < len(constituent_data.starts):
        start = constituent_data.starts[position]
        end = constituent_data.ends[position]

        if span.end <= start or span.end < end:
            break

        yield doc[start:end]
        position += 1

def get_child_spans(span):
    constituent_data, position = get_constituent(span)
    label_vocab = constituent_data.label_vocab
    doc = span.doc

    child_start_expected = span.start
    position += 1
    while position < len(constituent_data.starts):
        start = constituent_data.starts[position]
        end = constituent_data.ends[position]

        if span.end <= start or span.end < end:
            break

        if start == child_start_expected:
            yield doc[start:end]
            child_start_expected = end

        position += 1

def get_parent_span(span):
    constituent_data, position = get_constituent(span)
    label_vocab = constituent_data.label_vocab
    doc = span.doc
    sent = span.sent

    position -= 1
    while position >= 0:
        start = constituent_data.starts[position]
        end = constituent_data.ends[position]

        if start <= span.start and span.end <= end:
            return doc[start:end]
        if end < span.sent.start:
            break
        position -= 1

    return None

#%%

Span.set_extension('labels', getter=get_labels)
Span.set_extension('parse_string', getter=parse_string)
Span.set_extension('constituents', getter=get_subconstituents)
Span.set_extension('parent', getter=get_parent_span)
Span.set_extension('children', getter=get_child_spans)

Token.set_extension('labels', getter=lambda token: get_labels(token.doc[token.i:token.i+1]))
Token.set_extension('parse_string', getter=lambda token: parse_string(token.doc[token.i:token.i+1]))
Token.set_extension('parent', getter=lambda token: get_parent_span(token.doc[token.i:token.i+1]))



import json

BENEPAR_SERVER_INDEX = "https://kitaev.io/benepar_models/index.xml"

_downloader = None
def get_downloader():
    global _downloader
    if _downloader is None:
        import nltk.downloader
        _downloader = nltk.downloader.Downloader(server_index_url=BENEPAR_SERVER_INDEX)
    return _downloader

def download(*args, **kwargs):
    return get_downloader().download(*args, **kwargs)

def load_model(name):
    import nltk.data

    name_gz = "models/{}.gz".format(name)
    name_zip = "models/{}.zip".format(name)
    use_gz = False
    try:
        nltk.data.find(name_gz)
        use_gz = True
    except LookupError:
        pass

    try:
        if use_gz:
            return nltk.data.load(name_gz, format="raw")
        else:
            return {
                'meta': json.loads(nltk.data.load(name_zip + "/meta.json", format="text")),
                'model': nltk.data.load(name_zip + "/model.pb", format="raw"),
                'vocab': nltk.data.load(name_zip + "/vocab.txt", format="text", encoding="utf-8"),
                }
    except LookupError as e:
        arg = e.args[0].replace("nltk.download", "benepar.download")
    raise LookupError(arg)



import nltk
from nltk import Tree

from .base_parser import BaseParser, IS_PY2, STRING_TYPES, PTB_TOKEN_ESCAPE

TOKENIZER_LOOKUP = {
    'en': 'english',
    'de': 'german',
    'fr': 'french',
    'pl': 'polish',
    'sv': 'swedish',
}

class Parser(BaseParser):
    """
    Berkeley Neural Parser (benepar), integrated with NLTK.

    Sample usage:
    >>> parser = benepar.Parser("benepar_en")
    >>> parser.parse("The quick brown fox jumps over the lazy dog.")

    Note that the self-attentive parsing model is only tasked with constructing
    constituency parse trees from tokenized, untagged, single-sentence inputs.
    Any other elements of the NLP pipeline (i.e. sentence segmentation,
    tokenization, and part-of-speech tagging) will be done using the default
    NLTK models.
    """
    def __init__(self, name, batch_size=64):
        """
        Load a parsing model given a short model name (e.g. "benepar_en") or a
        filename on disk.

        name (str): Model name, or path to uncompressed TensorFlow model graph
        batch_size (int): Maximum number of sentences to process per batch
        """
        super(Parser, self).__init__(name, batch_size)
        self._tokenizer_lang = TOKENIZER_LOOKUP.get(self._language_code, None)

    def _make_nltk_tree(self, sentence, tags, score, p_i, p_j, p_label):
        # The optimized cython decoder implementation doesn't actually
        # generate trees, only scores and span indices. When converting to a
        # tree, we assume that the indices follow a preorder traversal.
        last_splits = []

        # Python 2 doesn't support "nonlocal", so wrap idx in a list
        idx_cell = [-1]
        def make_tree():
            idx_cell[0] += 1
            idx = idx_cell[0]
            i, j, label_idx = p_i[idx], p_j[idx], p_label[idx]
            label = self._label_vocab[label_idx]
            if (i + 1) >= j:
                if self._provides_tags:
                    word = sentence[i]
                    tag = self._tag_vocab[tags[i]]
                else:
                    word, tag = sentence[i]
                tag = PTB_TOKEN_ESCAPE.get(tag, tag)
                word = PTB_TOKEN_ESCAPE.get(word, word)
                tree = Tree(tag, [word])
                for sublabel in label[::-1]:
                    tree = Tree(sublabel, [tree])
                return [tree]
            else:
                left_trees = make_tree()
                right_trees = make_tree()
                children = left_trees + right_trees
                if label:
                    tree = Tree(label[-1], children)
                    for sublabel in reversed(label[:-1]):
                        tree = Tree(sublabel, [tree])
                    return [tree]
                else:
                    return children

        tree = make_tree()[0]
        tree.score = score

        return tree

    def _nltk_process_sents(self, sents):
        for sentence in sents:
            if isinstance(sentence, STRING_TYPES):
                if self._tokenizer_lang is None:
                    raise ValueError(
                        "No word tokenizer available for this language. "
                        "Please tokenize before calling the parser."
                        )
                sentence = nltk.word_tokenize(sentence, self._tokenizer_lang)

            if IS_PY2:
                sentence = [
                    word.decode('utf-8', 'ignore') if isinstance(word, str) else word
                    for word in sentence
                    ]

            if not self._provides_tags:
                sentence = nltk.pos_tag(sentence)
                yield [word for word, tag in sentence], sentence
            else:
                yield sentence, sentence

    def parse(self, sentence):
        """
        Parse a single sentence

        The argument "sentence" can be a list of tokens to be passed to the
        parser. It can also be a string, in which case the sentence will be
        tokenized using the default NLTK tokenizer.

        sentence (str or List[str]): sentence to parse

        Returns: nltk.Tree
        """
        return list(self.parse_sents([sentence]))[0]

    def parse_sents(self, sents):
        """
        Parse multiple sentences

        If "sents" is a string, it will be segmented into sentences using NLTK.
        Otherwise, each element of "sents" will be treated as a sentence.

        sents (str or Iterable[str] or Iterable[List[str]]): sentences to parse

        Returns: Iter[nltk.Tree]
        """
        if isinstance(sents, STRING_TYPES):
            if self._tokenizer_lang is None:
                raise ValueError(
                    "No tokenizer available for this language. "
                    "Please split into individual sentences and tokens "
                    "before calling the parser."
                    )
            sents = nltk.sent_tokenize(sents, self._tokenizer_lang)

        for parse_raw, tags_raw, sentence in self._batched_parsed_raw(self._nltk_process_sents(sents)):
            yield self._make_nltk_tree(sentence, tags_raw, *parse_raw)



import numpy as np
cimport numpy as np
from numpy cimport ndarray
cimport cython

ctypedef np.float32_t DTYPE_t

# @cython.boundscheck(False)
def decode(np.ndarray[DTYPE_t, ndim=3] label_scores_chart):
    cdef DTYPE_t NEG_INF = -np.inf
    cdef int sentence_len = label_scores_chart.shape[0] - 1

    cdef np.ndarray[DTYPE_t, ndim=2] value_chart = np.zeros((sentence_len+1, sentence_len+1), dtype=np.float32)
    cdef np.ndarray[int, ndim=2] split_idx_chart = np.zeros((sentence_len+1, sentence_len+1), dtype=np.int32)
    cdef np.ndarray[int, ndim=2] best_label_chart = np.zeros((sentence_len+1, sentence_len+1), dtype=np.int32)

    cdef int length
    cdef int left
    cdef int right

    cdef DTYPE_t label_score
    cdef int argmax_label_index

    cdef int best_split
    cdef int split_idx # Loop variable for splitting
    cdef DTYPE_t split_val # best so far
    cdef DTYPE_t max_split_val

    cdef int label_index_iter

    for length in range(1, sentence_len + 1):
        for left in range(0, sentence_len + 1 - length):
            right = left + length

            # We do argmax ourselves to make sure it compiles to pure C
            if length < sentence_len:
                argmax_label_index = 0
            else:
                # Not-a-span label is not allowed at the root of the tree
                argmax_label_index = 1

            label_score = label_scores_chart[left, right, argmax_label_index]
            for label_index_iter in range(1, label_scores_chart.shape[2]):
                if label_scores_chart[left, right, label_index_iter] > label_score:
                    argmax_label_index = label_index_iter
                    label_score = label_scores_chart[left, right, label_index_iter]
            best_label_chart[left, right] = argmax_label_index

            if length == 1:
                value_chart[left, right] = label_score
                continue

            best_split = left + 1
            split_val = NEG_INF
            for split_idx in range(left + 1, right):
                max_split_val = value_chart[left, split_idx] + value_chart[split_idx, right]
                if max_split_val > split_val:
                    split_val = max_split_val
                    best_split = split_idx

            value_chart[left, right] = label_score + value_chart[left, best_split] + value_chart[best_split, right]
            split_idx_chart[left, right] = best_split

    # Now we need to recover the tree by traversing the chart starting at the
    # root. This iterative implementation is faster than any of my attempts to
    # use helper functions and recursion

    # All fully binarized trees have the same number of nodes
    cdef int num_tree_nodes = 2 * sentence_len - 1
    cdef np.ndarray[int, ndim=1] included_i = np.empty(num_tree_nodes, dtype=np.int32)
    cdef np.ndarray[int, ndim=1] included_j = np.empty(num_tree_nodes, dtype=np.int32)
    cdef np.ndarray[int, ndim=1] included_label = np.empty(num_tree_nodes, dtype=np.int32)

    cdef int idx = 0
    cdef int stack_idx = 1
    # technically, the maximum stack depth is smaller than this
    cdef np.ndarray[int, ndim=1] stack_i = np.empty(num_tree_nodes + 5, dtype=np.int32)
    cdef np.ndarray[int, ndim=1] stack_j = np.empty(num_tree_nodes + 5, dtype=np.int32)
    stack_i[1] = 0
    stack_j[1] = sentence_len

    cdef int i, j, k
    while stack_idx > 0:
        i = stack_i[stack_idx]
        j = stack_j[stack_idx]
        stack_idx -= 1
        included_i[idx] = i
        included_j[idx] = j
        included_label[idx] = best_label_chart[i, j]
        idx += 1
        if i + 1 < j:
            k = split_idx_chart[i, j]
            stack_idx += 1
            stack_i[stack_idx] = k
            stack_j[stack_idx] = j
            stack_idx += 1
            stack_i[stack_idx] = i
            stack_j[stack_idx] = k

    cdef DTYPE_t score = value_chart[0, sentence_len]

    return score, included_i.astype(int), included_j.astype(int), included_label.astype(int)



import setuptools
import sys

try:
    from Cython.Build import cythonize
    USE_CYTHON = True
except ImportError:
    sys.exit("""Could not import Cython, which is required to build benepar extension modules.
Please install cython and numpy prior to installing benepar.""")

try:
    import numpy as np
except ImportError:
    sys.exit("""Could not import numpy, which is required to build the extension modules.
Please install cython and numpy prior to installing benepar.""")

with open("README.md", "r") as f:
    long_description = f.read()

extensions = cythonize("benepar/*.pyx")
for ext_module in extensions:
    ext_module.include_dirs.append(np.get_include())

setuptools.setup(
    name="benepar",
    version="0.1.1",
    author="Nikita Kitaev",
    author_email="kitaev@cs.berkeley.edu",
    description="Berkeley Neural Parser",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nikitakit/self-attentive-parser",
    packages=setuptools.find_packages(),
    package_data={'': ['*.pyx']},
    ext_modules = cythonize(extensions),
    classifiers=(
        'Programming Language :: Python :: 2.7',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Text Processing :: Linguistic",
    ),
    setup_requires = ["cython", "numpy"],
    install_requires = ["cython", "numpy", "nltk>=3.2"],
    extras_require={
        "cpu": ["tensorflow>=1.11.0"],
        "gpu": ["tensorflow-gpu>=1.11.0"],
        "spacy": ["spacy>=2.0.9"],
    },
)



import tensorflow as tf
import numpy as np
import os
import sys
import json
import codecs

from . import chart_decoder
from .downloader import load_model
from .bert_tokenization import BertTokenizer

#%%

IS_PY2 = sys.version_info < (3,0)

if IS_PY2:
    STRING_TYPES = (str, unicode)
else:
    STRING_TYPES = (str,)

ELMO_START_SENTENCE = 256
ELMO_STOP_SENTENCE = 257
ELMO_START_WORD = 258
ELMO_STOP_WORD = 259
ELMO_CHAR_PAD = 260

PTB_TOKEN_ESCAPE = {u"(": u"-LRB-",
    u")": u"-RRB-",
    u"{": u"-LCB-",
    u"}": u"-RCB-",
    u"[": u"-LSB-",
    u"]": u"-RSB-"}

BERT_TOKEN_MAPPING = {u"-LRB-": u"(",
    u"-RRB-": u")",
    u"-LCB-": u"{",
    u"-RCB-": u"}",
    u"-LSB-": u"[",
    u"-RSB-": u"]",
    u"``": u'"',
    u"''": u'"',
    u"`": u"'",
    u"\u201c": u'"',
    u"\u201d": u'"',
    u"\u2018": u"'",
    u"\u2019": u"'",
    u"\xab": u'"',
    u"\xbb": u'"',
    u"\u201e": u'"',
    u"\u2039": u"'",
    u"\u203a": u"'",
    }

# Label vocab is made immutable because it is potentially exposed to users
# through the spacy plugin
LABEL_VOCAB = ((),
 ('S',),
 ('PP',),
 ('NP',),
 ('PRN',),
 ('VP',),
 ('ADVP',),
 ('SBAR', 'S'),
 ('ADJP',),
 ('QP',),
 ('UCP',),
 ('S', 'VP'),
 ('SBAR',),
 ('WHNP',),
 ('SINV',),
 ('FRAG',),
 ('NAC',),
 ('WHADVP',),
 ('NP', 'QP'),
 ('PRT',),
 ('S', 'PP'),
 ('S', 'NP'),
 ('NX',),
 ('S', 'ADJP'),
 ('WHPP',),
 ('SBAR', 'S', 'VP'),
 ('SBAR', 'SINV'),
 ('SQ',),
 ('NP', 'NP'),
 ('SBARQ',),
 ('SQ', 'VP'),
 ('CONJP',),
 ('ADJP', 'QP'),
 ('FRAG', 'NP'),
 ('FRAG', 'ADJP'),
 ('WHADJP',),
 ('ADJP', 'ADJP'),
 ('FRAG', 'PP'),
 ('S', 'ADVP'),
 ('FRAG', 'SBAR'),
 ('PRN', 'S'),
 ('PRN', 'S', 'VP'),
 ('INTJ',),
 ('X',),
 ('NP', 'NP', 'NP'),
 ('FRAG', 'S', 'VP'),
 ('ADVP', 'ADVP'),
 ('RRC',),
 ('VP', 'PP'),
 ('VP', 'VP'),
 ('SBAR', 'FRAG'),
 ('ADVP', 'ADJP'),
 ('LST',),
 ('NP', 'NP', 'QP'),
 ('PRN', 'SBAR'),
 ('VP', 'S', 'VP'),
 ('S', 'UCP'),
 ('FRAG', 'WHNP'),
 ('NP', 'PP'),
 ('NP', 'SBAR', 'S', 'VP'),
 ('WHNP', 'QP'),
 ('VP', 'FRAG', 'ADJP'),
 ('FRAG', 'WHADVP'),
 ('NP', 'ADJP'),
 ('VP', 'SBAR'),
 ('NP', 'S', 'VP'),
 ('X', 'PP'),
 ('S', 'VP', 'VP'),
 ('S', 'VP', 'ADVP'),
 ('WHNP', 'WHNP'),
 ('NX', 'NX'),
 ('FRAG', 'ADVP'),
 ('FRAG', 'VP'),
 ('VP', 'ADVP'),
 ('SBAR', 'WHNP'),
 ('FRAG', 'SBARQ'),
 ('PP', 'PP'),
 ('PRN', 'PP'),
 ('VP', 'NP'),
 ('X', 'NP'),
 ('PRN', 'SINV'),
 ('NP', 'SBAR'),
 ('PP', 'NP'),
 ('NP', 'INTJ'),
 ('FRAG', 'INTJ'),
 ('X', 'VP'),
 ('PRN', 'NP'),
 ('FRAG', 'UCP'),
 ('NP', 'ADVP'),
 ('SBAR', 'SBARQ'),
 ('SBAR', 'SBAR', 'S'),
 ('SBARQ', 'WHADVP'),
 ('ADVP', 'PRT'),
 ('UCP', 'ADJP'),
 ('PRN', 'FRAG', 'WHADJP'),
 ('FRAG', 'S'),
 ('S', 'S'),
 ('FRAG', 'S', 'ADJP'),
 ('INTJ', 'S'),
 ('ADJP', 'NP'),
 ('X', 'ADVP'),
 ('FRAG', 'WHPP'),
 ('NP', 'FRAG'),
 ('NX', 'QP'),
 ('NP', 'S'),
 ('SBAR', 'WHADVP'),
 ('X', 'SBARQ'),
 ('NP', 'PRN'),
 ('NX', 'S', 'VP'),
 ('NX', 'S'),
 ('UCP', 'PP'),
 ('RRC', 'VP'),
 ('ADJP', 'ADVP'))

SENTENCE_MAX_LEN = 300
BERT_MAX_LEN = 512

#%%
class BaseParser(object):
    def __init__(self, name, batch_size=64):
        self._graph = tf.Graph()

        with self._graph.as_default():
            if isinstance(name, STRING_TYPES) and '/' not in name:
                model = load_model(name)
            elif not os.path.exists(name):
                raise Exception("Argument is neither a valid module name nor a path to an existing file/folder: {}".format(name))
            else:
                if not os.path.isdir(name):
                    with open(name, 'rb') as f:
                        model = f.read()
                else:
                    model = {}
                    with open(os.path.join(name, 'meta.json')) as f:
                        model['meta'] = json.load(f)
                    with open(os.path.join(name, 'model.pb'), 'rb') as f:
                        model['model'] = f.read()
                    with codecs.open(os.path.join(name, 'vocab.txt'), encoding='utf-8') as f:
                        model['vocab'] = f.read()

            if isinstance(model, dict):
                graph_def = tf.GraphDef.FromString(model['model'])
            else:
                graph_def = tf.GraphDef.FromString(model)
            tf.import_graph_def(graph_def, name='')

        self._sess = tf.Session(graph=self._graph)
        if not isinstance(model, dict):
            # Older model format (for ELMo-based models)
            self._chars = self._graph.get_tensor_by_name('chars:0')
            self._charts = self._graph.get_tensor_by_name('charts:0')
            self._label_vocab = LABEL_VOCAB
            self._language_code = 'en'
            self._provides_tags = False
            self._make_feed_dict = self._make_feed_dict_elmo
        else:
            # Newer model format (for BERT-based models)
            meta = model['meta']
            # Label vocab is made immutable because it is potentially exposed to
            # users through the spacy plugin
            self._label_vocab = tuple([tuple(label) for label in meta['label_vocab']])
            self._language_code = meta['language_code']
            self._provides_tags = meta['provides_tags']

            self._input_ids = self._graph.get_tensor_by_name('input_ids:0')
            self._word_end_mask = self._graph.get_tensor_by_name('word_end_mask:0')
            self._charts = self._graph.get_tensor_by_name('charts:0')
            if self._provides_tags:
                self._tag_vocab = meta['tag_vocab']
                self._tags = self._graph.get_tensor_by_name('tags:0')

            self._bert_tokenizer = BertTokenizer(
                model['vocab'], do_lower_case=meta['bert_do_lower_case'])
            self._make_feed_dict = self._make_feed_dict_bert

        self.batch_size = batch_size

    def _make_feed_dict_elmo(self, sentences):
        padded_len = max([len(sentence) + 2 for sentence in sentences])
        if padded_len > SENTENCE_MAX_LEN:
            raise ValueError("Sentence of length {} exceeds the maximum supported length of {}".format(
                padded_len - 2, SENTENCE_MAX_LEN - 2))

        all_chars = np.zeros((len(sentences), padded_len, 50), dtype=np.int32)

        for snum, sentence in enumerate(sentences):
            all_chars[snum, :len(sentence)+2,:] = ELMO_CHAR_PAD

            all_chars[snum, 0, 0] = ELMO_START_WORD
            all_chars[snum, 0, 1] = ELMO_START_SENTENCE
            all_chars[snum, 0, 2] = ELMO_STOP_WORD

            for i, word in enumerate(sentence):
                word = PTB_TOKEN_ESCAPE.get(word, word)
                if IS_PY2:
                    chars = [ELMO_START_WORD] + [ord(char) for char in word.encode('utf-8', 'ignore')[:(50-2)]] + [ELMO_STOP_WORD]
                else:
                    chars = [ELMO_START_WORD] + list(word.encode('utf-8', 'ignore')[:(50-2)]) + [ELMO_STOP_WORD]
                all_chars[snum, i+1, :len(chars)] = chars

            all_chars[snum, len(sentence)+1, 0] = ELMO_START_WORD
            all_chars[snum, len(sentence)+1, 1] = ELMO_STOP_SENTENCE
            all_chars[snum, len(sentence)+1, 2] = ELMO_STOP_WORD

            # Add 1; 0 is a reserved value for signaling words past the end of the
            # sentence, which we don't have because batch_size=1
            all_chars[snum, :len(sentence)+2,:] += 1

        return {self._chars: all_chars}

    def _make_feed_dict_bert(self, sentences):
        all_input_ids = np.zeros((len(sentences), BERT_MAX_LEN), dtype=int)
        all_word_end_mask = np.zeros((len(sentences), BERT_MAX_LEN), dtype=int)

        subword_max_len = 0
        for snum, sentence in enumerate(sentences):
            tokens = []
            word_end_mask = []

            tokens.append(u"[CLS]")
            word_end_mask.append(1)

            cleaned_words = []
            for word in sentence:
                word = BERT_TOKEN_MAPPING.get(word, word)
                # BERT is pre-trained with a tokenizer that doesn't split off
                # n't as its own token
                if word == u"n't" and cleaned_words:
                    cleaned_words[-1] = cleaned_words[-1] + u"n"
                    word = u"'t"
                cleaned_words.append(word)

            for word in cleaned_words:
                word_tokens = self._bert_tokenizer.tokenize(word)
                for _ in range(len(word_tokens)):
                    word_end_mask.append(0)
                word_end_mask[-1] = 1
                tokens.extend(word_tokens)
            tokens.append(u"[SEP]")
            word_end_mask.append(1)

            input_ids = self._bert_tokenizer.convert_tokens_to_ids(tokens)
            if len(sentence) + 2 > SENTENCE_MAX_LEN or len(input_ids) > BERT_MAX_LEN:
                raise ValueError("Sentence of length {} is too long to be parsed".format(
                    len(sentence)))

            subword_max_len = max(subword_max_len, len(input_ids))

            all_input_ids[snum, :len(input_ids)] = input_ids
            all_word_end_mask[snum, :len(word_end_mask)] = word_end_mask

        all_input_ids = all_input_ids[:, :subword_max_len]
        all_word_end_mask = all_word_end_mask[:, :subword_max_len]

        return {
            self._input_ids: all_input_ids,
            self._word_end_mask: all_word_end_mask
        }

    def _make_charts_and_tags(self, sentences):
        feed_dict = self._make_feed_dict(sentences)
        if self._provides_tags:
            charts_val, tags_val = self._sess.run((self._charts, self._tags), feed_dict)
        else:
            charts_val = self._sess.run(self._charts, feed_dict)
        for snum, sentence in enumerate(sentences):
            chart_size = len(sentence) + 1
            chart = charts_val[snum,:chart_size,:chart_size,:]
            if self._provides_tags:
                tags = tags_val[snum,1:chart_size]
            else:
                tags = None
            yield chart, tags

    def _batched_parsed_raw(self, sentence_data_pairs):
        batch_sentences = []
        batch_data = []
        for sentence, datum in sentence_data_pairs:
            batch_sentences.append(sentence)
            batch_data.append(datum)
            if len(batch_sentences) >= self.batch_size:
                for (chart_np, tags_np), datum in zip(self._make_charts_and_tags(batch_sentences), batch_data):
                    yield chart_decoder.decode(chart_np), tags_np, datum
                batch_sentences = []
                batch_data = []
        if batch_sentences:
            for (chart_np, tags_np), datum in zip(self._make_charts_and_tags(batch_sentences), batch_data):
                yield chart_decoder.decode(chart_np), tags_np, datum
